import operator

from typing import (
    TypeVar,
    Set,
    Hashable,
    Any,
    Callable,
    Type,
    Dict,
    List,
    Union as UnionType,
    Optional,
    Iterable,
    Iterator,
    MutableSet,
)

from odex.index import Index
from odex.optimize import Chain, Rule
from odex.parse import Parser
from odex.plan import Plan, Union, Intersect, ScanFilter, Filter, Planner, IndexLookup
from odex import condition as cond
from odex.condition import BinOp, UnaryOp, Attribute, Literal, Condition

T = TypeVar("T", bound=Hashable)

Attributes = Dict[str, Callable[[T, str], Any]]


class IndexedSet(MutableSet[T]):
    """
    Unordered, indexed collection of distinct, hashable objects.

    This is intended for efficient filtering of large sets by member attributes.

    Example:
        >>> from collections import namedtuple
        >>> from odex import HashIndex
        >>> X = namedtuple("X", ["a"])
        >>> iset = IndexedSet({X(a=1), X(a=2), X(a=3)}, indexes=[HashIndex("a")])
        >>> iset.filter("a = 2")
        {X(a=2)}

    Args:
        objs: Objects to initialize the set with
            Example: `{X(a=1), X(a=2)}`
        indexes: Attribute indexes for this set.
            Example: `[HashIndex("a")]`
        attrs: Extend members with extra attribute getters.
            Example: `{"a": lambda obj: obj.get_a()}`
        parser: Override the query parser
        planner: Override the query planner
        optimizer: Override the query optimizer
    """

    BINOPS = {
        cond.Add: operator.add,
        cond.Div: operator.truediv,
        cond.FloorDiv: operator.floordiv,
        cond.BitwiseAnd: operator.and_,
        cond.Xor: operator.xor,
        cond.BitwiseOr: operator.or_,
        cond.Pow: operator.pow,
        cond.Is: operator.is_,
        cond.Lshift: operator.lshift,
        cond.Mod: operator.mod,
        cond.Mul: operator.mul,
        cond.Rshift: operator.rshift,
        cond.Sub: operator.sub,
        cond.Lt: operator.lt,
        cond.Le: operator.le,
        cond.Gt: operator.gt,
        cond.Ge: operator.ge,
        cond.Eq: operator.eq,
        cond.Ne: operator.ne,
        cond.And: lambda l, r: l and r,
        cond.Or: lambda l, r: l or r,
        cond.In: lambda l, r: l in r,
    }
    UNARY_OPS = {
        cond.Not: operator.not_,
        cond.Invert: operator.invert,
    }

    def __init__(
        self,
        objs: Optional[Iterable[T]] = None,
        indexes: Optional[List[Index[T]]] = None,
        attrs: Optional[Attributes] = None,
        parser: Optional[Parser] = None,
        planner: Optional[Planner] = None,
        optimizer: Optional[Rule] = None,
    ):
        self.objs = objs if isinstance(objs, set) else set(objs) if objs else set()
        self.planner = planner or Planner()
        self.optimizer = optimizer or Chain()
        self.parser = parser or Parser()
        self.attrs = attrs or {}
        self.indexes = indexes or []
        self.update(self.objs)

        self.executors: Dict[Type[Plan], Callable[[Plan], Set[T]]] = {
            ScanFilter: lambda plan: {o for o in self.objs if self.match(plan.condition, o)},  # type: ignore
            Filter: lambda plan: {
                o
                for o in self.execute(plan.input)  # type: ignore
                if self.match(plan.condition, o)  # type: ignore
            },
            Union: lambda plan: set.union(*(self.execute(i) for i in plan.inputs)),  # type: ignore
            Intersect: lambda plan: set.intersection(*(self.execute(i) for i in plan.inputs)),  # type: ignore
            IndexLookup: lambda plan: plan.index.lookup(plan.value),  # type: ignore
        }

        def match_binop(op: Callable[[Any, Any], Any]) -> Callable[[BinOp, T], Any]:
            def matcher(condition: BinOp, obj: T) -> Any:
                return op(self.match(condition.left, obj), self.match(condition.right, obj))

            return matcher

        def match_unaryop(op: Callable[[Any], Any]) -> Callable[[UnaryOp, T], Any]:
            def matcher(condition: UnaryOp, obj: T) -> Any:
                return op(self.match(condition.operand, obj))

            return matcher

        self.matchers: Dict[Type[Condition], Callable[[Condition, T], Any]] = {
            Literal: lambda condition, obj: condition.value,  # type: ignore
            Attribute: lambda condition, obj: self.getattr(obj, condition.name),  # type: ignore
            **{klass: match_binop(op) for klass, op in self.BINOPS.items()},  # type: ignore
            **{klass: match_unaryop(op) for klass, op in self.UNARY_OPS.items()},  # type: ignore
        }

    def filter(self, condition: UnionType[Condition, str]) -> Set[T]:
        """
        Apply a logical expression to this set, returning a set of the matching members.

        Example:
            >>> iset = IndexedSet({X(a=1), X(a=2), X(a=3)}, indexes=[HashIndex("a")])
            >>> iset.filter("a = 2")
            {X(a=2)}
            >>> iset.filter(attr("a").eq(2))
            {X(a=2)}

        Args:
            condition: logical expression
        Returns:
            Set of matching members
        """
        plan = self.plan(condition)
        plan = self.optimize(plan)
        return self.execute(plan)

    def plan(self, condition: UnionType[Condition, str]) -> Plan:
        """
        Build a query plan from a condition.

        Args:
            condition: logical expression
        Returns:
            Query plan
        """
        if isinstance(condition, str):
            condition = self.parser.parse(condition)
        return self.planner.plan(condition)

    def optimize(self, plan: Plan) -> Plan:
        """
        Optimize a query plan.

        Args:
            plan: query plan
        Returns:
            Optimized query plan
        """
        return self.optimizer(plan, self)

    def execute(self, plan: Plan) -> Set[T]:
        """
        Execute a query plan.

        Args:
            plan: query plan
        Returns:
            Set of matching members
        """
        executor = self.executors.get(plan.__class__)
        if executor:
            return executor(plan)
        raise ValueError(f"Unsupported plan: {plan}")

    def match(self, condition: Condition, obj: T) -> Any:
        """
        Determine if `obj` matches the given `condition`.

        Args:
            condition: logical expression
            obj: object
        Returns:
            Usually a boolean, but this can return other types based on the condition
        """
        matcher = self.matchers.get(condition.__class__)
        if matcher:
            return matcher(condition, obj)

        raise ValueError(f"Unsupported condition: {condition}")

    def getattr(self, obj: T, item: str) -> Any:
        """Get the attribute `item` from `obj`"""
        return self.attrs.get(item, getattr)(obj, item)

    def add(self, obj: T) -> None:
        self.objs.add(obj)
        for index in self.indexes:
            index.add({obj}, self)

    def discard(self, obj: T) -> None:
        self.objs.discard(obj)
        for index in self.indexes:
            index.remove({obj}, self)

    def __contains__(self, x: Any) -> bool:
        return x in self.objs

    def __len__(self) -> int:
        return len(self.objs)

    def __iter__(self) -> Iterator[T]:
        for i in self.objs:
            yield i

    def update(self, objs: Set[T]) -> None:
        self.objs.update(objs)
        for index in self.indexes:
            index.add(objs, self)

    def difference_update(self, objs: Set[T]) -> None:
        self.objs.difference_update(objs)
        for index in self.indexes:
            index.remove(objs, self)



from typing import Callable, TypeVar, Dict, Type, cast, Optional

from odex import condition as cond
from odex.condition import Condition, Literal, Attribute, BinOp, UnaryOp
from sqlglot import Dialect, exp

B = TypeVar("B", bound=BinOp)
U = TypeVar("U", bound=UnaryOp)


class Converter:
    """Convert sqlglot Expressions to odex Conditions"""

    def __init__(self) -> None:
        self.converters: Dict[Type[exp.Expression], Callable[[exp.Expression], Condition]] = {
            exp.Add: lambda e: self._convert_binary(cond.Add, e),
            exp.Div: lambda e: self._convert_binary(cond.Div, e),
            exp.IntDiv: lambda e: self._convert_binary(cond.FloorDiv, e),
            exp.BitwiseAnd: lambda e: self._convert_binary(cond.BitwiseAnd, e),
            exp.BitwiseXor: lambda e: self._convert_binary(cond.Xor, e),
            exp.BitwiseOr: lambda e: self._convert_binary(cond.BitwiseOr, e),
            exp.Pow: lambda e: self._convert_binary(cond.Pow, e),
            exp.Is: lambda e: self._convert_binary(cond.Is, e),
            exp.BitwiseLeftShift: lambda e: self._convert_binary(cond.Lshift, e),
            exp.Mod: lambda e: self._convert_binary(cond.Mod, e),
            exp.Mul: lambda e: self._convert_binary(cond.Mul, e),
            exp.BitwiseRightShift: lambda e: self._convert_binary(cond.Rshift, e),
            exp.Sub: lambda e: self._convert_binary(cond.Sub, e),
            exp.LT: lambda e: self._convert_binary(cond.Lt, e),
            exp.LTE: lambda e: self._convert_binary(cond.Le, e),
            exp.GT: lambda e: self._convert_binary(cond.Gt, e),
            exp.GTE: lambda e: self._convert_binary(cond.Ge, e),
            exp.EQ: lambda e: self._convert_binary(cond.Eq, e),
            exp.NEQ: lambda e: self._convert_binary(cond.Ne, e),
            exp.And: lambda e: self._convert_binary(cond.And, e),
            exp.Or: lambda e: self._convert_binary(cond.Or, e),
            exp.Not: lambda e: self._convert_unary(cond.Not, e),
            exp.Literal: self._convert_literal,
            exp.Column: self._convert_column,
            exp.In: self._convert_in,
        }

    def convert(self, expression: exp.Expression) -> Condition:
        converter = self.converters.get(expression.__class__)
        if not converter:
            raise ValueError(f"Unsupported sqlglot Expression: {expression.__class__}")
        return converter(expression)

    def _convert_binary(self, condition_type: Type[B], expression: exp.Expression) -> B:
        expression = cast(exp.Binary, expression)
        return condition_type(
            self.convert(expression.left),
            self.convert(expression.right),
        )

    def _convert_unary(self, condition_type: Type[U], expression: exp.Expression) -> U:
        return condition_type(
            self.convert(expression.this),
        )

    def _convert_literal(self, expression: exp.Expression) -> Literal:
        expression = cast(exp.Literal, expression)
        if expression.is_string:
            return Literal(value=expression.this)
        if expression.is_int:
            return Literal(value=int(expression.this))
        if expression.is_number:
            return Literal(value=float(expression.this))
        raise ValueError(f"Unsupported sqlglot Literal: {expression}")

    def _convert_column(self, expression: exp.Expression) -> Attribute:
        expression = cast(exp.Column, expression)
        return Attribute(name=expression.name)

    def _convert_in(self, expression: exp.Expression) -> cond.In:
        expression = cast(exp.In, expression)
        field = expression.args.get("field")
        if field:
            return cond.In(left=self.convert(expression.this), right=Attribute(name=field.name))
        raise ValueError(f"Unsupported sqlglot In: {expression}")


class Parser:
    """Parse SQL-like logical expressions into odex Conditions"""

    def __init__(self, dialect: Optional[Dialect] = None, converter: Optional[Converter] = None):
        self.dialect = dialect or Dialect()
        self.converter = converter or Converter()

    def parse(self, expression: str) -> Condition:
        ast = self.dialect.parse_into(exp.Condition, expression)[0]
        if not ast:
            raise ValueError(f"Failed to parse expression: {expression}")
        return self.converter.convert(ast)



"""
Query plans. These make up the nodes of a query plan.
"""

from abc import abstractmethod
from dataclasses import dataclass
from copy import deepcopy
from typing import List, Any, Callable, TYPE_CHECKING
from typing_extensions import Protocol

from odex.condition import Condition, And, Or

if TYPE_CHECKING:
    from odex.index import Index


Transformer = Callable[["Plan"], "Plan"]


class Plan(Protocol):
    """Base class for all query plan nodes"""

    def __str__(self) -> str:
        return self.to_s()

    @abstractmethod
    def to_s(self, depth: int = 0) -> str:
        """Render this plan as a string"""

    def transform_inputs(self, transformer: Transformer) -> None:
        """
        Recursively transform any child plans

        Args:
            transformer: function that transforms plan nodes and returns a new node
        Returns:
            nothing - this should happen in-place
        """
        return None

    def transform(self, transformer: Transformer) -> "Plan":
        """
        Recursively transform this plan using the `transformer` function.

        Args:
            transformer: function that transforms this plan node and returns a new node
        Returns:
            new plan
        """
        self.transform_inputs(transformer)
        return transformer(self)


@dataclass
class ScanFilter(Plan):
    """Return all objects in the collection, filtering with `condition`"""

    condition: Condition

    def to_s(self, depth=0):
        return f"ScanFilter: {self.condition}"


@dataclass
class Filter(Plan):
    """Return all objects return by `input`, filtering with `condition`"""

    condition: Condition
    input: Plan

    def to_s(self, depth=0):
        indent = "  " * depth
        return f"Filter: {self.condition}\n{indent}  - {self.input.to_s(depth + 1)}"

    def transform_inputs(self, transformer: Transformer) -> None:
        self.input = self.input.transform(transformer)


@dataclass
class SetOp(Plan):
    """Base class for set operations"""

    inputs: List[Plan]

    def to_s(self, depth=0):
        indent = "  " * depth
        inputs = "\n".join([f"{indent}  - {i.to_s(depth+1)}" for i in self.inputs])
        return f"{self.__class__.__name__}\n{inputs}"

    def transform_inputs(self, transformer: Transformer) -> None:
        self.inputs = [i.transform(transformer) for i in self.inputs]


@dataclass
class Intersect(SetOp):
    """Return the intersection of all `inputs`"""


@dataclass
class Union(SetOp):
    """Return the union of all `inputs`"""


@dataclass
class IndexLookup(Plan):
    """Return objects by looking up `value` in `index`"""

    index: "Index"
    value: Any
    cost: int

    def to_s(self, depth=0):
        return f"IndexLookup: {self.index} = {self.value}"

    def __deepcopy__(self, memodict):
        return IndexLookup(index=self.index, value=deepcopy(self.value), cost=self.cost)


class Planner:
    def plan(self, condition: Condition) -> Plan:
        """
        Convert a syntax tree into a query plan.

        Args:
            condition: logical expression syntax tree
        Returns:
            query plan
        """
        if isinstance(condition, And):
            return Intersect(inputs=[self.plan(condition.left), self.plan(condition.right)])
        if isinstance(condition, Or):
            return Union(inputs=[self.plan(condition.left), self.plan(condition.right)])
        return ScanFilter(condition=condition)



from typing import Dict, Type, Callable, Sequence, cast
from typing_extensions import Protocol

from odex.condition import and_
from odex.context import Context
from odex.plan import Plan, SetOp, Intersect, Filter, ScanFilter, Union, IndexLookup


class Rule(Protocol):
    def __call__(self, plan: Plan, ctx: Context) -> Plan: ...


class TransformerWithContext:
    """
    Implementation of the Plan Transformer protocol, including Context
    """

    def __init__(self, ctx: Context, func: Callable[[Plan, Context], Plan]):
        self.ctx = ctx
        self.func = func

    def __call__(self, plan: Plan) -> Plan:
        return self.func(plan, self.ctx)


class TransformerRule(Rule):
    """Abstract class for optimizer rules that call `Plan.transform`"""

    def __call__(self, plan: Plan, ctx: Context) -> Plan:
        return plan.transform(TransformerWithContext(ctx, self.transform))

    def transform(self, plan: Plan, ctx: Context) -> Plan:
        return plan


class MergeSetOps(TransformerRule):
    """Merge nested set operations into a single operation"""

    def transform(self, plan: Plan, ctx: Context) -> Plan:
        if isinstance(plan, SetOp):
            new_inputs = []
            for input in plan.inputs:
                if isinstance(input, plan.__class__):
                    new_inputs.extend(input.inputs)
                else:
                    new_inputs.append(input)
            plan.inputs = new_inputs
        return plan


class UseIndex(TransformerRule):
    """Replace scans with index lookups"""

    def transform(self, plan: Plan, ctx: Context) -> Plan:
        if isinstance(plan, ScanFilter):
            best_plan: Plan = plan
            best_cost = len(ctx.objs)
            for idx in ctx.indexes:
                match = idx.match(plan.condition)
                if match:
                    if best_cost is None or match.cost <= best_cost:
                        best_plan, best_cost = match, match.cost
            return best_plan

        return plan


class CombineFilters(TransformerRule):
    """Combine multiple filters into one"""

    def transform(self, plan: Plan, ctx: Context) -> Plan:
        if isinstance(plan, Intersect):
            others = []
            filters = []
            for i in plan.inputs:
                if isinstance(i, ScanFilter):
                    filters.append(i)
                else:
                    others.append(i)

            if filters:
                combined = and_(*(f.condition for f in filters))
                if others:
                    base = others[0] if len(others) == 1 else Intersect(inputs=others)
                    return Filter(
                        condition=combined,
                        input=base,
                    )
                else:
                    return ScanFilter(
                        condition=combined,
                    )
        return plan


class OrderIntersects(Rule):
    """Reorder intersections so that the plans with the smallest cost are first"""

    def __init__(self) -> None:
        self.estimators: Dict[Type[Plan], Callable[[Plan, Context], int]] = {
            ScanFilter: lambda plan, ctx: len(ctx.objs),
            Filter: lambda plan, ctx: self._estimate(cast(Filter, plan).input, ctx),
            IndexLookup: lambda plan, ctx: cast(IndexLookup, plan).cost,
            Union: lambda plan, ctx: max(self._estimate(i, ctx) for i in cast(Union, plan).inputs),
            Intersect: self._estimate_intersect,
        }

    def __call__(self, plan: Plan, ctx: Context) -> Plan:
        self._estimate(plan, ctx)
        return plan

    def _estimate(self, plan: Plan, ctx: Context) -> int:
        estimator = self.estimators.get(plan.__class__)
        return estimator(plan, ctx) if estimator else len(ctx.objs)

    def _estimate_intersect(self, plan: Plan, ctx: Context) -> int:
        plan = cast(Intersect, plan)
        costs = sorted(((self._estimate(i, ctx), i) for i in plan.inputs), key=lambda t: t[0])
        plan.inputs = [i[1] for i in costs]
        return costs[0][0]


class Chain(Rule):
    """Chain multiple rules together"""

    DEFAULT_RULES = (MergeSetOps(), UseIndex(), CombineFilters(), OrderIntersects())

    def __init__(self, rules: Sequence[Rule] = DEFAULT_RULES):
        self.rules = list(rules)

    def __call__(self, plan: Plan, ctx: Context) -> Plan:
        for rule in self.rules:
            plan = rule(plan, ctx)
            # print(rule)
            # print(plan)
        return plan



from abc import abstractmethod
from typing import Generic, TypeVar, Set, Any, Optional, Iterable, Dict
from typing_extensions import Protocol

from odex.condition import Condition, Eq, Literal, Attribute, In
from odex.context import Context
from odex.plan import IndexLookup

T = TypeVar("T")


class Index(Protocol[T]):
    @abstractmethod
    def add(self, objs: Set[T], ctx: Context[T]) -> None:
        """Add `objs` to the index"""

    @abstractmethod
    def remove(self, objs: Set[T], ctx: Context[T]) -> None:
        """Remove `objs` from the index"""

    @abstractmethod
    def match(self, condition: Condition) -> "Optional[IndexLookup]":
        """
        Determine if this index can serve the given `condition`.

        Args:
            condition: logical expression
        Returns:
            `None` if this index can't serve the condition.
            `IndexLoop` plan if it can.
        """

    @abstractmethod
    def lookup(self, value: Any) -> Set[T]:
        """
        Get members from the index.

        Args:
            value: Attribute value to lookup
        Returns:
            Result set
        """


_NotFound = object()


class HashIndex(Generic[T], Index[T]):
    """
    Hash table index.

    This maps unique object attribute values to sets of objects.

    This matches equality expressions, e.g. `a = 1`

    Args:
        attr: name of the attribute to index
    """

    def __init__(self, attr: str):
        self.attr = attr
        self.idx: Dict[Any, Set[T]] = {}

    def add(self, objs: Set[T], ctx: Context[T]) -> None:
        for o in objs:
            for val in self._extract_values(o, ctx):
                self.idx.setdefault(val, set()).add(o)

    def remove(self, objs: Set[T], ctx: Context[T]) -> None:
        for o in objs:
            for val in self._extract_values(o, ctx):
                self.idx.setdefault(val, set()).remove(o)

    def lookup(self, value: Any) -> Set[T]:
        return self.idx.get(value) or set()

    def match(self, condition: Condition) -> Optional[IndexLookup]:
        value = self._match(condition)
        if value is _NotFound:
            return None
        objs = self.idx.get(value) or set()
        return IndexLookup(index=self, cost=len(objs), value=value)

    def _extract_values(self, obj: T, ctx: Context[T]) -> Iterable[Any]:
        yield ctx.getattr(obj, self.attr)

    def _match(self, condition: Condition) -> Any:
        if isinstance(condition, Eq):
            l, r = condition.left, condition.right
            if isinstance(l, Attribute) and l.name == self.attr and isinstance(r, Literal):
                return r.value
            elif isinstance(r, Attribute) and r.name == self.attr and isinstance(l, Literal):
                return l.value
        return _NotFound

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.attr})"


class MultiHashIndex(Generic[T], HashIndex[T]):
    """
    Same as a `HashIndex`, except this assumes the attribute is a collection of values.

    This matches IN expressions, e.g. `1 in a`
    """

    def _extract_values(self, obj: T, ctx: Context[T]) -> Iterable[Any]:
        for val in ctx.getattr(obj, self.attr):
            yield val

    def _match(self, condition: Condition) -> Any:
        if isinstance(condition, In):
            member, container = condition.left, condition.right
            if isinstance(member, Literal) and isinstance(container, Attribute):
                return member.value
        return _NotFound



"""
Logical conditions. These make up the nodes of a syntax tree.
"""

from dataclasses import dataclass
from functools import reduce

from typing import Any, ClassVar


class Condition:
    """Base class for all conditions"""

    def and_(self, *others: Any) -> "Condition":
        return and_(self, *others)

    def or_(self, *others: Any) -> "Condition":
        return or_(self, *others)

    def add(self, other: Any) -> "Condition":
        return add(self, other)

    def div(self, other: Any) -> "Condition":
        return div(self, other)

    def floordiv(self, other: Any) -> "Condition":
        return floordiv(self, other)

    def bitwise_and(self, other: Any) -> "Condition":
        return bitwise_and(self, other)

    def xor(self, other: Any) -> "Condition":
        return xor(self, other)

    def bitwise_or(self, other: Any) -> "Condition":
        return bitwise_or(self, other)

    def pow(self, other: Any) -> "Condition":
        return pow(self, other)

    def is_(self, other: Any) -> "Condition":
        return is_(self, other)

    def lshift(self, other: Any) -> "Condition":
        return lshift(self, other)

    def mod(self, other: Any) -> "Condition":
        return mod(self, other)

    def mul(self, other: Any) -> "Condition":
        return mul(self, other)

    def rshift(self, other: Any) -> "Condition":
        return rshift(self, other)

    def sub(self, other: Any) -> "Condition":
        return sub(self, other)

    def lt(self, other: Any) -> "Condition":
        return lt(self, other)

    def le(self, other: Any) -> "Condition":
        return le(self, other)

    def gt(self, other: Any) -> "Condition":
        return gt(self, other)

    def ge(self, other: Any) -> "Condition":
        return ge(self, other)

    def eq(self, other: Any) -> "Condition":
        return eq(self, other)

    def ne(self, other: Any) -> "Condition":
        return ne(self, other)

    def in_(self, other: Any) -> "Condition":
        return in_(self, other)

    def not_(self) -> "Condition":
        return not_(self)

    def invert(self) -> "Condition":
        return invert(self)


@dataclass
class Attribute(Condition):
    """Name of an object attribute"""

    name: str

    def __str__(self) -> str:
        return self.name


@dataclass
class Literal(Condition):
    """Any Python value"""

    value: Any

    def __str__(self) -> str:
        return str(self.value)


@dataclass
class BinOp(Condition):
    """Abstract class for binary operators"""

    left: Condition
    right: Condition
    SYMBOL: ClassVar[str] = ""

    def __str__(self) -> str:
        return f"{self.left} {self.SYMBOL} {self.right}"


@dataclass
class Add(BinOp):
    SYMBOL = "+"


@dataclass
class Div(BinOp):
    SYMBOL = "/"


@dataclass
class FloorDiv(BinOp):
    SYMBOL = "//"


@dataclass
class BitwiseAnd(BinOp):
    SYMBOL = "&"


@dataclass
class Xor(BinOp):
    SYMBOL = "^"


@dataclass
class BitwiseOr(BinOp):
    SYMBOL = "|"


@dataclass
class Pow(BinOp):
    SYMBOL = "**"


@dataclass
class Is(BinOp):
    SYMBOL = "is"


@dataclass
class Lshift(BinOp):
    SYMBOL = "<<"


@dataclass
class Mod(BinOp):
    SYMBOL = "%"


@dataclass
class Mul(BinOp):
    SYMBOL = "*"


@dataclass
class Rshift(BinOp):
    SYMBOL = ">>"


@dataclass
class Sub(BinOp):
    SYMBOL = "-"


@dataclass
class Lt(BinOp):
    SYMBOL = "<"


@dataclass
class Le(BinOp):
    SYMBOL = "<="


@dataclass
class Gt(BinOp):
    SYMBOL = ">"


@dataclass
class Ge(BinOp):
    SYMBOL = ">="


@dataclass
class Eq(BinOp):
    SYMBOL = "="


@dataclass
class Ne(BinOp):
    SYMBOL = "!="


@dataclass
class Or(BinOp):
    SYMBOL = "OR"


@dataclass
class And(BinOp):
    SYMBOL = "AND"


@dataclass
class In(BinOp):
    SYMBOL = "IN"


@dataclass
class UnaryOp(Condition):
    """Abstract class for unary operators"""

    operand: Condition
    SYMBOL: ClassVar[str] = ""

    def __str__(self) -> str:
        return f"{self.SYMBOL} {self.operand}"


@dataclass
class Not(UnaryOp):
    SYMBOL = "NOT"


@dataclass
class Invert(UnaryOp):
    SYMBOL = "~"


# Just some sugar for the fluent interface
literal = Literal
attr = Attribute


def ensure_condition(val: Any) -> Condition:
    if isinstance(val, Condition):
        return val
    return Literal(val)


def and_(*conditions: Any) -> Condition:
    """Logically AND multiple conditions"""
    return reduce(lambda l, r: And(l, r), [ensure_condition(c) for c in conditions])


def or_(*conditions: Any) -> Condition:
    """Logically OR multiple conditions"""
    return reduce(lambda l, r: Or(l, r), [ensure_condition(c) for c in conditions])


def add(left: Any, right: Any) -> Condition:
    return Add(ensure_condition(left), ensure_condition(right))


def div(left: Any, right: Any) -> Condition:
    return Div(ensure_condition(left), ensure_condition(right))


def floordiv(left: Any, right: Any) -> Condition:
    return FloorDiv(ensure_condition(left), ensure_condition(right))


def bitwise_and(left: Any, right: Any) -> Condition:
    return BitwiseAnd(ensure_condition(left), ensure_condition(right))


def xor(left: Any, right: Any) -> Condition:
    return Xor(ensure_condition(left), ensure_condition(right))


def bitwise_or(left: Any, right: Any) -> Condition:
    return BitwiseOr(ensure_condition(left), ensure_condition(right))


def pow(left: Any, right: Any) -> Condition:
    return Pow(ensure_condition(left), ensure_condition(right))


def is_(left: Any, right: Any) -> Condition:
    return Is(ensure_condition(left), ensure_condition(right))


def lshift(left: Any, right: Any) -> Condition:
    return Lshift(ensure_condition(left), ensure_condition(right))


def mod(left: Any, right: Any) -> Condition:
    return Mod(ensure_condition(left), ensure_condition(right))


def mul(left: Any, right: Any) -> Condition:
    return Mul(ensure_condition(left), ensure_condition(right))


def rshift(left: Any, right: Any) -> Condition:
    return Rshift(ensure_condition(left), ensure_condition(right))


def sub(left: Any, right: Any) -> Condition:
    return Sub(ensure_condition(left), ensure_condition(right))


def lt(left: Any, right: Any) -> Condition:
    return Lt(ensure_condition(left), ensure_condition(right))


def le(left: Any, right: Any) -> Condition:
    return Le(ensure_condition(left), ensure_condition(right))


def gt(left: Any, right: Any) -> Condition:
    return Gt(ensure_condition(left), ensure_condition(right))


def ge(left: Any, right: Any) -> Condition:
    return Ge(ensure_condition(left), ensure_condition(right))


def eq(left: Any, right: Any) -> Condition:
    return Eq(ensure_condition(left), ensure_condition(right))


def ne(left: Any, right: Any) -> Condition:
    return Ne(ensure_condition(left), ensure_condition(right))


def in_(left: Any, right: Any) -> Condition:
    return In(ensure_condition(left), ensure_condition(right))


def not_(operand: Any) -> Condition:
    return Not(ensure_condition(operand))


def invert(operand: Any) -> Condition:
    return Invert(ensure_condition(operand))



from odex.set import IndexedSet as IndexedSet
from odex.index import HashIndex as HashIndex, MultiHashIndex as MultiHashIndex
from odex.condition import (
    literal as literal,
    attr as attr,
    and_ as and_,
    or_ as or_,
    add as add,
    div as div,
    floordiv as floordiv,
    bitwise_and as bitwise_and,
    xor as xor,
    bitwise_or as bitwise_or,
    pow as pow,
    is_ as is_,
    lshift as lshift,
    mod as mod,
    mul as mul,
    rshift as rshift,
    sub as sub,
    lt as lt,
    le as le,
    gt as gt,
    ge as ge,
    eq as eq,
    ne as ne,
    in_ as in_,
    not_ as not_,
    invert as invert,
)



from dataclasses import dataclass, field

from typing import Any, TYPE_CHECKING, Set, TypeVar, List
from typing_extensions import Protocol

if TYPE_CHECKING:
    from odex.index import Index
    from odex.set import Attributes

T = TypeVar("T")


class Context(Protocol[T]):
    """Interface for filter context, so `IndexedSet` can pass context to optimizers"""

    indexes: "List[Index]"
    objs: Set[T]
    attrs: "Attributes"

    def getattr(self, obj: T, item: str) -> Any: ...


@dataclass
class SimpleContext(Context[T]):
    """Context as a dataclass. Intended for testing."""

    indexes: "List[Index]" = field(default_factory=list)
    objs: Set[T] = field(default_factory=set)
    attrs: "Attributes" = field(default_factory=dict)

    def getattr(self, obj: T, item: str) -> Any:
        return getattr(obj, item)



from typing import Generic, TypeVar

T = TypeVar("T")


class Container(Generic[T]):
    """
    Proxy object that wraps another object.

    This is intended for a convenience to wrap objects that aren't hashable, e.g. `dict`s.
    Container objects are hashed by `id`.

    Example:
        >>> from odex.container import Container
        >>> container = Container({"a": 1})
        >>> container.a
        1
        >>> hash(Container({"a": 1})) != hash(container)
        True
    """

    def __init__(self, obj: T):
        self.obj = obj

    def __getattr__(self, item):
        try:
            return getattr(self.obj, item)
        except AttributeError:
            return self.obj[item]
